import pytest

from lamin_logger._map_synonyms import explode_aggregated_column_to_expand, map_synonyms


@pytest.fixture(scope="module")
def genes():
    import pandas as pd

    gene_symbols = ["A1CF", "A1BG", "FANCD1", "FANCD20"]

    records = [
        {
            "symbol": "BRCA1",
            "synonyms": "PPP1R53|RNF53|FANCS|BRCC1",
        },
        {
            "symbol": "A1BG",
            "synonyms": "",
        },
        {
            "symbol": "BRCA2",
            "synonyms": "FAD|FAD1|BRCC2|FANCD1|FACD|FANCD|XRCC11",
        },
        {
            "symbol": "A1CF",
            "synonyms": "ACF|ACF64|APOBEC1CF|ACF65|ASP",
        },
        {
            "symbol": "PDCD1",
            "synonyms": "hSLE1|PD-1|PD1|SLEB2|CD279",
        },
        {
            "symbol": "PDCD1",
            "synonyms": "hSLE1|PD-1|PD1|SLEB2|CD279",
        },
    ]

    df = pd.DataFrame.from_records(records)

    return gene_symbols, df


def test_map_synonyms(genes):
    gene_symbols, df = genes

    mapping = map_synonyms(
        df=df, identifiers=gene_symbols, field="symbol", return_mapper=False
    )
    expected_synonym_mapping = ["A1CF", "A1BG", "BRCA2", "FANCD20"]
    assert mapping == expected_synonym_mapping


def test_map_synonyms_return_mapper(genes):
    gene_symbols, df = genes

    mapping = map_synonyms(
        df=df, identifiers=gene_symbols, field="symbol", return_mapper=True
    )

    expected_synonym_mapping = {"FANCD1": "BRCA2"}

    assert mapping == expected_synonym_mapping


def test_unsupported_field(genes):
    gene_symbols, df = genes
    with pytest.raises(KeyError):
        map_synonyms(df=df, identifiers=gene_symbols, field="name", return_mapper=False)


def test_explode_aggregated_column_to_expand(genes):
    _, df = genes
    with pytest.raises(AssertionError):
        explode_aggregated_column_to_expand(
            df=df, aggregated_col="synonyms", target_col="synonyms"
        )

    res = explode_aggregated_column_to_expand(
        df=df, aggregated_col="synonyms", target_col="symbol"
    )
    assert res.index.name == "synonyms"
    assert res.shape == (27, 1)



import re
from collections import namedtuple
from typing import Any, Dict, Iterable, List, Optional, Tuple


def _append_records_to_list(df_dict: Dict, value: str, record) -> None:
    """Append unique records to a list."""
    values_list = df_dict[value]
    if not isinstance(values_list, list):
        values_list = [values_list]
    values_set = set(values_list)
    values_set.add(record)
    df_dict[value] = list(values_set)


def _create_df_dict(
    df: Any = None,
    field: Optional[str] = None,
    records: Optional[List] = None,
    values: Optional[List] = None,
    tuple_name: Optional[str] = None,
) -> Dict:
    """Create a dict with {lookup key: records in namedtuple}.

    Value is a list of namedtuples if multiple records match the same key.
    """
    if df is not None:
        records = df.itertuples(index=False, name=tuple_name)
        values = df[field]
    df_dict: Dict = {}  # a dict of namedtuples as records and values as keys
    for i, row in enumerate(records):  # type:ignore
        value = values[i]  # type:ignore
        if not isinstance(value, str):
            continue
        if value in df_dict:
            _append_records_to_list(df_dict=df_dict, value=value, record=row)
        else:
            df_dict[value] = row
    return df_dict


class Lookup:
    """Lookup object with dot and [] access."""

    # removed DataFrame type annotation to speed up import time
    def __init__(
        self,
        field: Optional[str] = None,
        tuple_name="MyTuple",
        prefix: str = "bt",
        df: Any = None,
        values: Optional[Iterable] = None,
        records: Optional[List] = None,
    ) -> None:
        self._tuple_name = tuple_name
        if df is not None:
            values = df[field]
        self._df_dict = _create_df_dict(
            df=df,
            field=field,
            records=records,
            values=values,  # type:ignore
            tuple_name=self._tuple_name,
        )
        lkeys = self._to_lookup_keys(values=values, prefix=prefix)  # type:ignore
        self._lookup_dict = self._create_lookup_dict(lkeys=lkeys, df_dict=self._df_dict)

    def _to_lookup_keys(self, values: Iterable, prefix: str) -> Dict:
        """Convert a list of strings to tab-completion allowed formats.

        Returns:
            {lookup_key: value_or_values}
        """
        lkeys: Dict = {}
        for value in list(values):
            if not isinstance(value, str):
                continue
            # replace any special character with _
            lkey = re.sub("[^0-9a-zA-Z_]+", "_", str(value)).lower()
            if lkey == "":  # empty strings are skipped
                continue
            if not lkey[0].isalpha():  # must start with a letter
                lkey = f"{prefix.lower()}_{lkey}"

            if lkey in lkeys:
                # if multiple values have the same lookup key
                # put the values into a list
                _append_records_to_list(df_dict=lkeys, value=lkey, record=value)
            else:
                lkeys[lkey] = value
        return lkeys

    def _create_lookup_dict(self, lkeys: Dict, df_dict: Dict) -> Dict:
        lkey_dict: Dict = {}  # a dict of namedtuples as records and lookup keys as keys
        for lkey, values in lkeys.items():
            if isinstance(values, list):
                combined_list = []
                for v in values:
                    records = df_dict.get(v)
                    if isinstance(records, list):
                        combined_list += records
                    else:
                        combined_list.append(records)
                lkey_dict[lkey] = combined_list
            else:
                lkey_dict[lkey] = df_dict.get(values)

        return lkey_dict

    def dict(self) -> Dict:
        """Dictionary of the lookup."""
        return self._df_dict

    def lookup(self) -> Tuple:
        """Lookup records with dot access."""
        keys: List = list(self._lookup_dict.keys()) + ["dict"]
        MyTuple = namedtuple("Lookup", keys)  # type:ignore
        return MyTuple(**self._lookup_dict, dict=self.dict)  # type:ignore



from typing import Any, Dict, Iterable, List, Union


def map_synonyms(
    df: Any,
    identifiers: Iterable,
    field: str,
    *,
    synonyms_field: str = "synonyms",
    sep: str = "|",
    return_mapper: bool = False,
) -> Union[Dict[str, str], List[str]]:
    """Maps input identifiers against a concatenated synonyms column.

    Args:
        identifiers: Identifiers that will be mapped against an field.
        field: The field representing the identifiers.
        return_mapper: If True, returns {identifiers : <mapped field values>}.

    Returns:
        - A list of mapped field values if return_mapper is False.
        - A dictionary of mapped values with mappable identifiers as keys
            and values mapped to field as values if return_mapper is True.
    """
    if field not in df.columns:
        raise KeyError(
            f"field '{field}' is invalid! Available fields are: {list(df.columns)}"
        )
    if synonyms_field not in df.columns:
        raise KeyError(
            f"synonyms_field '{synonyms_field}' is invalid! Available fields"
            f" are: {list(df.columns)}"
        )
    if field == synonyms_field:
        raise KeyError("synonyms_field must be different from field!")

    alias_map = explode_aggregated_column_to_expand(
        df,
        aggregated_col=synonyms_field,
        target_col=field,
        sep=sep,
    )[field]

    if return_mapper:
        mapped_dict = {
            item: alias_map.get(item)
            for item in identifiers
            if alias_map.get(item) is not None and alias_map.get(item) != item
        }
        return mapped_dict
    else:
        mapped_list = [alias_map.get(item, item) for item in identifiers]
        return mapped_list


def explode_aggregated_column_to_expand(
    df: Any,
    aggregated_col: str,
    target_col=None,
    sep: str = "|",
) -> Any:
    """Explode values from an aggregated DataFrame column to expand a target column.

    Args:
        df: A DataFrame containing the aggregated_col and target_col.
        aggregated_col: The name of the aggregated column
        target_col: the name of the target column
                    If None, use the index as the target column
        sep: Splits all values of the aggregated_col by this separator.

    Returns:
        a DataFrame index by the split values from the aggregated column;
        the target column is aggregated so that the new index is unique.
    """
    import pandas as pd

    if target_col is None:
        # take the index as the target column
        if df.index.name is None:
            target_col = df.index.name = "index"
        else:
            target_col = df.index.name
    if aggregated_col == target_col:
        raise AssertionError("synonyms and target column can't be the same!")
    try:
        df = df.reset_index()[[aggregated_col, target_col]].copy()
    except KeyError:
        raise KeyError(f"{aggregated_col} field is not found!")

    # explode the values from the aggregated cells into new rows
    df[aggregated_col] = df[aggregated_col].str.split(sep)
    exploded_df = df.explode(aggregated_col)

    # if any values in the aggregated column is already in the target col
    # sets those values of the aggregated column to None
    exploded_df.loc[
        exploded_df[aggregated_col].isin(exploded_df[target_col]), aggregated_col
    ] = None

    # set the values aggregated_col equal the target_col if None before concat
    exploded_df[aggregated_col] = exploded_df[aggregated_col].fillna(
        exploded_df[target_col]
    )

    # append the additional values in the target column to the df
    add_values = exploded_df[
        ~exploded_df[target_col].isin(exploded_df[aggregated_col])
    ][target_col].unique()
    add_df = pd.DataFrame(data={target_col: add_values, aggregated_col: add_values})

    # aggregate the target column so that the new index (aggregated column) is unique
    df_concat = pd.concat([exploded_df, add_df])
    df_concat = df_concat.astype(str)
    df_concat = df_concat.groupby(aggregated_col).agg(sep.join)

    return df_concat



from typing import Any, Union

from lamin_logger import logger


def search(
    df: Any,
    string: str,
    field: str = "name",
    synonyms_field: Union[str, None] = "synonyms",
    case_sensitive: bool = True,
    return_ranked_results: bool = False,
    synonyms_sep: str = "|",
    tuple_name: str = "SearchResult",
) -> Any:
    """Search a given string against a field.

    Args:
        string: The input string to match against the field ontology values.
        field: The field against which the input string is matching.
        synonyms_field: Also map against in the synonyms
            If None, no mapping against synonyms.
        case_sensitive: Whether the match is case sensitive.
        return_ranked_results: If True, return all entries ranked by matching ratios.

    Returns:
        Best match record of the input string.
    """
    import pandas as pd

    from ._map_synonyms import explode_aggregated_column_to_expand

    def _fuzz_ratio(string: str, iterable: pd.Series, case_sensitive: bool = True):
        from rapidfuzz import fuzz, utils

        if case_sensitive:
            processor = None
        else:
            processor = utils.default_process
        return iterable.apply(lambda x: fuzz.ratio(string, x, processor=processor))

    if (synonyms_field in df.columns) and (synonyms_field != field):
        df_exp = explode_aggregated_column_to_expand(
            df,
            aggregated_col=synonyms_field,  # type:ignore
            target_col=field,
            sep=synonyms_sep,
        ).reset_index()
        target_column = synonyms_field
    else:
        if synonyms_field == field:
            logger.warning(
                "Input field is the same as synonyms field, skipping synonyms matching"
            )
        df_exp = df.copy()
        target_column = field

    df_exp["__ratio__"] = _fuzz_ratio(
        string=string, iterable=df_exp[target_column], case_sensitive=case_sensitive
    )
    df_exp_grouped = (
        df_exp.groupby(field).max().sort_values("__ratio__", ascending=False)
    )
    df_exp_grouped = df_exp_grouped[df_exp_grouped.index.isin(df[field])]
    df_scored = df.set_index(field).loc[df_exp_grouped.index]
    df_scored["__ratio__"] = df_exp_grouped["__ratio__"]

    if return_ranked_results:
        return df_scored.sort_values("__ratio__", ascending=False)
    else:
        res = df_scored[df_scored["__ratio__"] == df_scored["__ratio__"].max()]
        res = res.drop(columns=["__ratio__"])
        res_records = list(res.reset_index().itertuples(index=False, name=tuple_name))
        if len(res_records) == 1:
            return res_records[0]
        elif len(res_records) > 1:
            return res_records
        else:
            return None



from pathlib import Path

import nbproject_test as test


def test_notebooks():
    # assuming this is in the tests folder
    docs_folder = Path(__file__).parents[1] / "docs/"

    for check_folder in docs_folder.glob("./**"):
        # these are the notebook testpaths
        if not str(check_folder).endswith(("guides", "tutorials")):
            continue
        test.execute_notebooks(check_folder, write=True)



import pandas as pd

from lamin_logger._lookup import Lookup


def test_lookup():
    df = pd.DataFrame(
        {
            "name": ["Sample 1", "Sample 1", "sample 1", "1 sample"],
            "meta1": ["metadata~1", "metadata~1~1", "metadata~1~1~1", "1 metadata"],
        }
    )

    inst = Lookup(df=df, field="name", tuple_name="TestTuple", prefix="prefix")
    lookup = inst.lookup()

    assert len(lookup.sample_1) == 3
    assert isinstance(lookup.sample_1, list)
    lookup_sample_1_dicts = [i._asdict() for i in lookup.sample_1]
    assert {"name": "Sample 1", "meta1": "metadata~1~1"} in lookup_sample_1_dicts
    assert {"name": "Sample 1", "meta1": "metadata~1"} in lookup_sample_1_dicts
    assert {"name": "sample 1", "meta1": "metadata~1~1~1"} in lookup_sample_1_dicts

    assert lookup.prefix_1_sample._asdict() == {
        "name": "1 sample",
        "meta1": "1 metadata",
    }

    lookup_dict = lookup.dict()
    assert len(lookup_dict) == 3
    assert isinstance(lookup_dict["Sample 1"], list)
    assert len(lookup_dict["Sample 1"]) == 2
    assert isinstance(lookup_dict["sample 1"], tuple)
    assert lookup_dict["1 sample"]._asdict() == {
        "name": "1 sample",
        "meta1": "1 metadata",
    }

    assert lookup.__class__.__name__ == "Lookup"
    assert lookup.prefix_1_sample.__class__.__name__ == "TestTuple"



# Parts of this class are from the Scanpy equivalent, see license below

# BSD 3-Clause License

# Copyright (c) 2017 F. Alexander Wolf, P. Angerer, Theis Lab
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:

# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.

# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.

# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Logging and Profiling."""
import logging
import platform
import sys
from datetime import datetime, timedelta, timezone
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING, getLevelName
from typing import Optional

HINT = 15
DOWNLOAD = 21
SUCCESS = 25
logging.addLevelName(HINT, "HINT")
logging.addLevelName(DOWNLOAD, "DOWNLOAD")
logging.addLevelName(SUCCESS, "SUCCESS")

VERBOSITY_TO_LOGLEVEL = {
    0: "ERROR",
    1: "WARNING",
    2: "INFO",
    3: "HINT",
    4: "DEBUG",
}


LEVEL_TO_ICONS = {
    40: "❌",  # error
    30: "🔶",  # warning
    25: "✅",  # success
    21: "💾",  # download
    20: "💬",  # info
    15: "💡",  # hint
    10: "🐛",  # debug
}


class RootLogger(logging.RootLogger):
    def __init__(self, level="INFO"):
        super().__init__(level)
        self.propagate = False
        self._verbosity: int = 1
        RootLogger.manager = logging.Manager(self)

    def log(  # type: ignore
        self,
        level: int,
        msg: str,
        *,
        extra: Optional[dict] = None,
        time: datetime = None,
        deep: Optional[str] = None,
    ) -> datetime:
        """Log message with level and return current time.

        Args:
            level: Logging level.
            msg: Message to display.
            time: A time in the past. If this is passed, the time difference from then
                to now is appended to `msg` as ` (HH:MM:SS)`.
                If `msg` contains `{time_passed}`, the time difference is instead
                inserted at that position.
            deep: If the current verbosity is higher than the log function’s level,
                this gets displayed as well
            extra: Additional values you can specify in `msg` like `{time_passed}`.
        """
        now = datetime.now(timezone.utc)
        time_passed: timedelta = None if time is None else now - time  # type: ignore
        extra = {
            **(extra or {}),
            "deep": deep
            if getLevelName(VERBOSITY_TO_LOGLEVEL[self._verbosity]) < level
            else None,
            "time_passed": time_passed,
        }
        super().log(level, msg, extra=extra)
        return now

    def critical(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(CRITICAL, msg, time=time, deep=deep, extra=extra)

    def error(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(ERROR, msg, time=time, deep=deep, extra=extra)

    def warning(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(WARNING, msg, time=time, deep=deep, extra=extra)

    def success(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(SUCCESS, msg, time=time, deep=deep, extra=extra)

    def info(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(INFO, msg, time=time, deep=deep, extra=extra)

    def download(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(DOWNLOAD, msg, time=time, deep=deep, extra=extra)

    def hint(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(HINT, msg, time=time, deep=deep, extra=extra)

    def debug(self, msg, *, time=None, deep=None, extra=None) -> datetime:  # type: ignore  # noqa
        return self.log(DEBUG, msg, time=time, deep=deep, extra=extra)


class _LogFormatter(logging.Formatter):
    def __init__(
        self, fmt="{levelname}: {message}", datefmt="%Y-%m-%d %H:%M", style="{"
    ):
        super().__init__(fmt, datefmt, style)

    def base_format(self, record: logging.LogRecord):
        if platform.system() == "Windows":
            return f"{record.levelname}:" + " {message}"
        else:
            return f"{LEVEL_TO_ICONS[record.levelno]}" + " {message}"

    def format(self, record: logging.LogRecord):
        format_orig = self._style._fmt
        self._style._fmt = self.base_format(record)
        if record.time_passed:  # type: ignore
            # strip microseconds
            if record.time_passed.microseconds:  # type: ignore
                record.time_passed = timedelta(  # type: ignore
                    seconds=int(record.time_passed.total_seconds())  # type: ignore
                )
            if "{time_passed}" in record.msg:
                record.msg = record.msg.replace(
                    "{time_passed}", str(record.time_passed)  # type: ignore
                )
            else:
                self._style._fmt += " ({time_passed})"
        if record.deep:  # type: ignore
            record.msg = f"{record.msg}: {record.deep}"  # type: ignore
        result = logging.Formatter.format(self, record)
        self._style._fmt = format_orig
        return result


logger = RootLogger()


def set_handler(logger):
    h = logging.StreamHandler(stream=sys.stdout)
    h.setFormatter(_LogFormatter())
    h.setLevel(logger.level)
    if len(logger.handlers) == 1:
        logger.removeHandler(logger.handlers[0])
    elif len(logger.handlers) > 1:
        raise RuntimeError("Lamin's root logger somehow got more than one handler")
    logger.addHandler(h)


set_handler(logger)


def set_log_level(logger, level: int):
    logger.setLevel(level)
    (h,) = logger.handlers  # can only be 1
    h.setLevel(level)


# this also sets it for the handler
RootLogger.set_level = set_log_level  # type: ignore


def set_verbosity(logger, verbosity: int):
    if verbosity not in VERBOSITY_TO_LOGLEVEL:
        raise ValueError(
            f"verbosity needs to be one of {set(VERBOSITY_TO_LOGLEVEL.keys())}"
        )
    logger.set_level(VERBOSITY_TO_LOGLEVEL[verbosity])
    logger._verbosity = verbosity


RootLogger.set_verbosity = set_verbosity  # type: ignore



"""Logging setup.

Import the package::

   import lamin_logger

This is the complete API reference:

.. autosummary::
   :toctree: .

   colors
"""

__version__ = "0.5.2"

from ._core import colors  # noqa
from ._logger import logger
from ._python_version import py_version_warning  # noqa



import laminci
import nox

nox.options.default_venv_backend = "none"


@nox.session
def lint(session: nox.Session) -> None:
    laminci.nox.run_pre_commit(session)


@nox.session
def build(session):
    session.run(*"pip install .[dev]".split())
    laminci.nox.run_pytest(session, coverage=False)
    laminci.nox.build_docs(session)



import pytest

from lamin_logger._search import search


@pytest.fixture(scope="module")
def df():
    import pandas as pd

    records = [
        {
            "ontology_id": "CL:0000084",
            "name": "T cell",
            "synonyms": "T-cell|T lymphocyte|T-lymphocyte",
            "children": ["CL:0000798", "CL:0002420", "CL:0002419", "CL:0000789"],
        },
        {
            "ontology_id": "CL:0000236",
            "name": "B cell",
            "synonyms": "B lymphocyte|B-lymphocyte|B-cell",
            "children": ["CL:0009114", "CL:0001201"],
        },
        {
            "ontology_id": "CL:0000696",
            "name": "PP cell",
            "synonyms": "type F enteroendocrine cell",
            "children": ["CL:0002680"],
        },
        {
            "ontology_id": "CL:0002072",
            "name": "nodal myocyte",
            "synonyms": "cardiac pacemaker cell|myocytus nodalis|P cell",
            "children": ["CL:1000409", "CL:1000410"],
        },
    ]
    return pd.DataFrame.from_records(records)


def test_search_name_default(df):
    res = search(df=df, string="T cells", tuple_name="MyTuple")
    assert res.name == "T cell"
    assert res.ontology_id == "CL:0000084"
    assert res.children == ["CL:0000798", "CL:0002420", "CL:0002419", "CL:0000789"]
    assert res.synonyms == "T-cell|T lymphocyte|T-lymphocyte"
    assert res.__class__.__name__ == "MyTuple"


def test_search_synonyms(df):
    res = search(df=df, string="P cells")
    assert res.name == "nodal myocyte"

    res = search(df=df, synonyms_field=None, string="P cells")
    assert res.name == "PP cell"


def test_search_return_df(df):
    res = search(df=df, string="P cells", return_ranked_results=True)
    assert res.shape == (4, 4)
    assert res.iloc[0].name == "nodal myocyte"


def test_search_return_tie_results(df):
    res = search(df=df, string="A cell", synonyms_field=None)
    assert len(res) == 2


def test_search_non_default_field(df):
    res = search(df=df, string="type F enteroendocrine", field="synonyms")
    res.synonyms == "type F enteroendocrine cell"
